/* Trithorpe interpreter C API for UNIX */

extern "C" {
#include "Tapi.h"
}

#include "trithorpe.h"
#include "compiler.h"

#include <stdio.h>
#include <malloc.h>

//typedef	    struct TriTerp   Tterp;
//typedef	    void     Tputchar(int charValue);
//typedef	    int		Terr;
//typedef	    char*	Tnta;   // null-terminated string allocation

//int	Tverbose;

extern "C" {

Tterp*	Tnew( )
{
    Tterp* T= new TriTerp();

    // Putc() hack
    T->home.aStore( tri('P','u','t','c'), tri("tcod putc nil ret") );
    T->home.aStore( tri('P','u','t','s'), tri("tcod puts nil ret") );
    T->home.aStore( tri('A','b','o','r'), tri("tcod abor nil ret") );
    T->home.aStore( tri('P','u','t','U'), tri("tcod putup nil ret") );
    T->home.aStore( tri('G','e','t','U'), tri("tcod getup nil ret") );

    return T;
}

void	Tdelete( Tterp* T )
{
    delete T;
}


void*	Talloc( int size )
{
    return calloc( size, 1 );
}

void 	Tfree( const void* p )
{
    free( (void*) p );
}

char* 	Tdup( const char* string, int len )
{
    if ( len<0 ) len= StrLen( string );
    char* p= (char*) Talloc( len+1 );
    for (int i=0; i<len; i++) p[i]= string[i];
    p[len]= '\0';
    return p;
}



void 	Tset( Tterp* T, const char* path, const char* value )
{
    int len= StrLen(path);
    tri key= NIL;
    if (len>0) {
	key.setTCodeItemFromString(path, 0, len);
    }
    tri val= Str(value);
    T->home.asTab()->store( key, val );
}

Tnta 	Tget( Tterp* T, const char* path )
{
    int len= StrLen(path);
    tri key= NIL;
    if (len>0) {
	key.setTCodeItemFromString(path, 0, len);
    }
    return StrDupIfPrintableAsciiObj( NIL << T->home.asTab()->recall( key ) );
}


Tnta	Teval( Tterp* T, const char* script )
{
    T->e= NIL;
    Error *error= NULL;
    tri libTokens= tri::NewCodVec( script );
    tri libCode("tcod");
    Compiler libpiler( libTokens, libCode );

    Expr* list= libpiler.parseExprList( &error );
    if ( error ) { T->e= error->getError(); say << "\n*** ERROR in parse: " << error << "\n";  delete error; return 0; }

    if (list) { T->e= list->emitGet(); delete list; } 
    if ( T->e != NIL ) { say << "\n*** ERROR IN emitGet: " << T->e << "\n";  return 0; }
    T->e= T->testEval( libCode, NIL, NIL );
    if ( T->e != NIL ) { say << "\n*** ERROR IN testEval:" << T->e << "\n";  return 0; }

    return StrDupIfPrintableAsciiObj( NIL << T->pop() );
}

Tnta	Terrmsg( Tterp* T )
{
    if ( T->e == NIL ) {
	return 0;
    } else {
	return StrDupIfPrintableAsciiObj( NIL << T->e );
    }
}

//////////// Internally Used


char* nullPointer;

void Tputc( int c ) { fputc( c, stderr ); }
void Tbomb( const char* s ) { fprintf( stderr, "Trithorpe BOMBING: <%s>\n", s ); nullPointer[1]= 1; }

// END 

}
