class Compiler;

class Expr {

public:

    tri	    what;  // in leiu of class object

    Compiler* piler;

    Expr(Compiler* c) : piler(c) {}
    virtual ~Expr();
    virtual err beginPut(); // abort()
    virtual err finishPut(); // abort()
    virtual err emitGet() = 0;
    virtual err emitCall(); // abort()
    virtual bool isError();
    virtual err getError();
    IFD( virtual void print() = 0; )
    void out(const tri& x);
    err emitAsABlock();
};

class ConsExpr : public Expr {
    Expr*	head;
    ConsExpr*	tail;

public:
    ConsExpr(Compiler*, Expr* _head, ConsExpr* _tail);
    ~ConsExpr();
    err emitGet();
    err emitGetEach();
    IFD( void print(); )
    int len();
};

class Builtin : public Expr {


    tri	    code;
    Expr    *a1;
    Expr    *a2;
    Expr    *a3;
    Expr    *a4;
    int	    nresults;
    int	    special;

public:
    Builtin(Compiler*, tri _code, Expr* _a1=0, Expr* _a2=0, Expr* _a3=0, Expr* _a4=0, int _nresults= 1, int _special= 0 );
    ~Builtin();
    err emitGet();
    IFD( void print(); )

    const static int THIRD_FIRST = 3;
};

class Verbatim : public Expr {
    int numverbs;
    tri	verb1;
    tri	verb2;

public:
    Verbatim(Compiler*, int _numverbs, tri _verb1, tri _verb2= NIL);
    ~Verbatim();
    err emitGet();
    IFD( void print(); )
};


class Lit : public Expr {
protected:
    tri	lit;

public:
    Lit(Compiler*, tri _lit);
    ~Lit();
    err emitGet();
    IFD( void print(); )
};

class Error : public Lit {
    Expr*	inner;

public:
    Error(Compiler*, err _err, Expr* _Inner= NULL);
    ~Error();
    err emitGet();
    virtual bool isError();
    virtual err getError();
    IFD( void print(); )
};

class LocalVar : public Expr {
    int	    index;

public:
    LocalVar(Compiler*, int _index);
    ~LocalVar();
    err beginPut();
    err finishPut();
    err emitGet();
    IFD( void print(); )
};

class Slot : public Expr {
    Expr*	    table;
    tri		    key;

public:
    Slot(Compiler*, Expr* _table, tri _key);
    ~Slot();
    err beginPut();
    err finishPut();
    err emitGet();
    err emitCall();
    IFD( void print(); )
};

class Subscript : public Expr {
    Expr*	    table;
    Expr*	    index;

public:
    Subscript(Compiler*, Expr* _table, Expr* _index);
    ~Subscript();
    err beginPut();
    err finishPut();
    err emitGet();
    err emitCall();
    IFD( void print(); )
};

class Assign : public Expr {
    Expr*	    lhs;
    Expr*	    rhs;

public:
    Assign(Compiler*, Expr* _lhs, Expr* _rhs);
    ~Assign();
    err emitGet();
    IFD( void print(); )
};

class Binop : public Expr {
    tri	    op;
    Expr*   one;
    Expr*   two;

public:
    Binop(Compiler*, tri _op, Expr* _a1, Expr* _a2);
    ~Binop();
    err emitGet();
    IFD( void print(); )
};

class IfExpr : public Expr {
    Expr*   cond;
    Expr*   thenBlock;
    Expr*   elseBlock;

public:
    IfExpr(Compiler*, Expr* _cond,  Expr* _thenBlock, Expr* _elseBlock );
    ~IfExpr();
    err emitGet();
    IFD( void print(); )
};

class LoopExpr : public Expr {
    Expr*   body;

public:
    LoopExpr(Compiler*, Expr* _body );
    ~LoopExpr();
    err emitGet();
    IFD( void print(); )
};

class Call : public Expr {
    Expr*	one;
    ConsExpr*   argList;

public:
    Call(Compiler*, Expr* _a1, ConsExpr* _argList);
    ~Call();
    err emitGet();
    IFD( void print(); )
};

struct ExprPair {
	err	e;  // for the error status
	Expr*	x;  // for the data

	ExprPair( err _e, Expr* _x ) : e(_e) , x(_x) {}
	ExprPair( const ExprPair& copyMe ) : e(copyMe.e) , x(copyMe.x) {}
	ExprPair& operator = ( const ExprPair& copyMe ) { e= copyMe.e; x= copyMe.x; return *this; }
	~ExprPair() {}
};
	
struct Pair {
	err	e;  // for the error status
	tri	x;  // for the data

	Pair( ) : e() , x() {}
	Pair( err _e, tri _x ) : e(_e) , x(_x) {}
	Pair( const Pair& copyMe ) : e(copyMe.e) , x(copyMe.x) {}
	Pair& operator = ( const Pair& copyMe ) { e= copyMe.e; x= copyMe.x; return *this; }
	~Pair() {}
};
	

class Compiler {

    tri	    src;
    int	    isize;
    int	    inext;

public:
    tri	    dest;

    tri	    numLocalVars;
    tri	    localVars;

    Compiler( tri source, tri dest );

    Expr* parseSuffices( Expr* stem );
    Expr* parseExpr( );
    Expr* parseTerm( );
    Expr* parseBasic( );
    Expr* parseFun( );
    ConsExpr* parseExprList( Error** errorOut );
    Expr* parseBlock( );
    Expr* parseIf( );
    Expr* parseLoop( );
    Expr* parseBuiltin( tri code, int nargs, int nresults= 1, int special=0 );

    err outputAsABlock( Expr* );

    void output(const tri& x);
    tri cur() { return src.aGet(inext); }
    void advance() { 
	if (verbose) say << "{*" << cur() << "*}";
	++inext; 
    }
    bool done() { return inext>=isize; }

    bool useLocal( ) { return localVars != NIL; }
    int addLocal( tri varName );
    int findLocal( tri varName );  // -1 means not found

};

// END
