#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "trithorpe.h"
#include "compiler.h"

TriTerp* T;

void showStack() { T->dumpStack(); }

#define MAX_SRC 999999


tri 
evalstring(TriTerp &terp, const char* script )
{
    Error *error= NULL;
    err e;
    tri libTokens= tri::NewCodVec( script );
    tri libCode("tcod");
    Compiler libpiler( libTokens, libCode );
    Expr* list= libpiler.parseExprList( &error );
    if ( error ) { say << "\n*** ERROR " << error << "\n";  Bomb("trimain: compile script returns error"); }
    if (list) { e= list->emitGet(); delete list; }
    e= terp.testEval( libCode, NIL, NIL );
    if ( e != tri() ) { say << "\n*** ERROR " << e << "\n";  Bomb("trimain: eval returns error"); }
    if (error) delete error;
    return NIL;
}

tri
evalfile(TriTerp &terp, const char* filename )
{
    char lib[ MAX_SRC ];

    int libfd= SYSopen( filename, O_RDONLY );
    if ( libfd < 0 ) { say << "\n*** CANNOT OPEN "<< filename <<"\n"; Bomb("trimain: cannot open file"); }
    int libcc= SYSread( libfd, lib, sizeof lib );
    if ( libcc < 0 ) { say << "\n*** CANNOT READ "<< filename <<"\n"; Bomb("trimain: cannot read file"); }
    lib[libcc]= '\0';
    SYSclose( libfd );

    return evalstring( terp, lib );
}


int 
trimain(int argc, const char* argv[])
{
    char buf[ MAX_SRC ];
    int cc= SYSread( 0, buf, sizeof buf );
    if ( cc < 0 ) { say << "\n*** CANNOT READ stdin\n"; Bomb("trimain: cannot read stdin"); }
    buf[cc]= '\0';

    Error *error= NULL;
    TriTerp terp;
    T= &terp; // for gdb

    // Argc
    terp.home.aStore( tri('A','r','g','c'), tri(argc) );

    // argv
    for (int i=0; i<argc; i++) {
	terp.home.aPut( i, Str( argv[i] ) );
    }

    // Putc() hack
    terp.home.aStore( tri('P','u','t','c'), tri("tcod putc nil ret") );
    terp.home.aStore( tri('P','u','t','s'), tri("tcod puts nil ret") );
    terp.home.aStore( tri('A','b','o','r'), tri("tcod abor nil ret") );
    terp.home.aStore( tri('P','u','t','U'), tri("tcod putup nil ret") );
    terp.home.aStore( tri('G','e','t','U'), tri("tcod getup nil ret") );

    ////// first the lib.tri

    evalfile( terp, "lib.tri" );

    /////// now the stdin program

    err e;
    tri program= tri::NewCodVec(buf);
    tri code("tcod");

#ifndef MINIMAL
    char* VERBOSE =  getenv("VERBOSE");
    verbose= atoi( VERBOSE ? VERBOSE : "0" );
#endif

    Compiler piler( program, code );
    
    Expr* list= piler.parseExprList( &error );
    if ( error ) { say << "\n*** ERROR " << error << "\n";  Bomb("trimain: eval stdin returns error"); }

    if (list) { e= list->emitGet(); delete list; }

    if (verbose) dumpVec( code, "stdin" );

    e= terp.testEval( code, NIL, NIL );
    if ( e != tri() ) { say << "\n*** ERROR " << e << "\n";  Bomb("trimain: eval stdin returns error"); }

    if (verbose) dumpVec( terp.home, "Home" );

    return 0;
}
